#ifndef  __GRF_KERNEL__
#define __GRF_KERNEL__

#ifndef _D_INTRND_
#define _D_INTRND_

inline __device__ UINT intrnd (UINT seed) // 1<=seed<m
{
    UINT const a = 16807; //ie 7**5
    //UINT const m = 16777215; //ie 2**24-1
    //seed = (long(seed * a)) % m;
    seed = (UINT) __umul24(seed, a);
    return seed;
}
#endif /*_D_INTRND_*/

//one block per thread for updating
__global__ void d_Update(BOOL *g_bufUpdate, BOOL *g_bufMatrix, UINT *g_bufCount)
{
    UINT tid = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
    if (tid < NUM_PEERS * NUM_BLOCKS)
    {
        g_bufMatrix[tid] += g_bufUpdate[tid];
        if (g_bufUpdate[tid])
            atomicAdd(&g_bufCount[tid % NUM_BLOCKS], 1);
        g_bufUpdate[tid] = 0;
    }
}

/* ==================== data block per thread ================== */
#ifdef GRF_DATABLOCK_PER_THREAD
//multiple blocks per thread for finding min
__global__ void d_p2pGRF(USHORT *g_neighList, BOOL *g_bufMatrix, UINT *g_bufCount, 
    BOOL *g_bufUpdate, UINT *g_doneCount, UINT seed)
{
    __shared__ USHORT receivePeer, sendPeer;
    __shared__ UINT   blk2Send;
    __shared__ UINT   addrStartSrc, addrStartDest;
    __shared__ UINT   iAmDone;
    __shared__ UINT   min;
    __shared__ UINT   iteration;

    UINT              tid      = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
    UINT              blkIndex;
     
    //printf("tid : %d, \n", tid);
    if (tid < NUM_PEERS * blockDim.x)
    {
        //blkIndex = tid % NUM_BLOCKS;
        //printf("tid : %d, \n", tid);
        
        // Peer Operations
        if (threadIdx.x == 0) // NUM_BLOCKS = a multiple of BLK_SZ
        {
            blk2Send = NUM_BLOCKS;
            iAmDone = TRUE;
            sendPeer      = blockIdx.x; 
            // Select a random target peer for transmission
            UINT tmp      = intrnd(seed+sendPeer) % g_neighList[sendPeer * NUM_PEERS]; 
            //UINT tmp = 0;
            receivePeer   = g_neighList[sendPeer*NUM_PEERS + tmp + 1];
            addrStartSrc  = __umul24(sendPeer, NUM_BLOCKS);
            addrStartDest = __umul24(receivePeer, NUM_BLOCKS);
            //printf("tid : %d, sendPeer %d, receivePeer %d\n", tid, sendPeer, receivePeer);
            min = NUM_PEERS + 1;
            iteration = 0;
        }
        
        __syncthreads();
        
        while (iteration * blockDim.x < NUM_BLOCKS)
        {
            blkIndex = iteration * blockDim.x  + threadIdx.x;
            if (blkIndex < NUM_BLOCKS)
            {
                __syncthreads();
                // Block Operations: find the rarest block              
                if (g_bufMatrix[addrStartSrc + blkIndex] == MATRIX_TRUE)
                {
                    if(g_bufMatrix[addrStartDest + blkIndex] == MATRIX_FALSE)
                    {
                        atomicMin(&min, g_bufCount[blkIndex]);
                        //printf("tid : %d, min %d\n", tid, min); 
                    }
                }
                else
                {
                    iAmDone = FALSE;
                }    
                __syncthreads();

                if (g_bufCount[blkIndex] == min && 
                    g_bufMatrix[addrStartSrc + blkIndex] > g_bufMatrix[addrStartDest + blkIndex])
                {   

                    atomicExch(&blk2Send, blkIndex);
                    //printf("tid : %d, blk2Send: %d\n", tid, blk2Send);  
                }
                __syncthreads();
            }        
            if (threadIdx.x == 0)
                iteration ++;
                
            __syncthreads();
        }
        
        __syncthreads();
        //Peer operations
        if (threadIdx.x == 1)
        {
            if (blk2Send < NUM_BLOCKS)
            {     
                //printf("tid : %d, Peer %d sends block %d to Peer %d\n", tid, sendPeer, blk2Send receivePeer);
                g_bufUpdate[addrStartDest + blk2Send] = 1; 
            }
        }
        if (threadIdx.x == 2)
        {
            //increase global done count
            if (iAmDone == TRUE) atomicAdd(g_doneCount, 1);
        }
    }
}

#else
/* ==================== Peer Per Thread =======================*/
__global__ void d_p2pGRF(USHORT *g_neighList, BOOL *g_bufMatrix, UINT *g_bufCount, 
    BOOL *g_bufUpdate, UINT *g_doneCount, UINT seed)
{
    //__shared__ USHORT s_blkDiffList[BLK_SZ][SMEM_DIFF_BLK_MAX];
    USHORT            receiver;
    UINT              blk2Send;
    UINT              index, count, tmp;
    UINT              addrStartSrc, addrStartDest;
    UINT              iAmDone  = FALSE;
    UINT              tid      = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
    UINT              min = NUM_PEERS + 1;
    
    if(tid < NUM_PEERS)
    {
        // Select a random target peer for transmission
#ifdef RANDOM_SELECTION
        index         = __umul24(tid, NUM_PEERS); // just reuse index to save registers
        tmp           = intrnd(seed+tid) % g_neighList[index]; 
        receiver   = g_neighList[index + tmp + 1];
#else
        receiver   = (tid+1) % NUM_PEERS;
#endif
    
        count         = 0;
        iAmDone       = TRUE;
        addrStartSrc  = __umul24(tid, NUM_BLOCKS);
        addrStartDest = __umul24(receiver, NUM_BLOCKS);
        
        //printf("##start## Transmitter %d -> Receiver %d    ", tid, receiver);
        //Find the rarest block to send
        for (index = 0; index < NUM_BLOCKS /*&& (iAmDone == TRUE || count < SMEM_DIFF_BLK_MAX)*/; index++)
        {
            //printf("start tid %d: target %d, index %d, src %d, target %d \n", 
            //tid, receiver, index, s_blkListSrc[threadIdx.x][index], g_bufCount[receiver*NUM_BLOCKS+index]);
            if (g_bufMatrix[addrStartSrc + index] == MATRIX_TRUE )
            {
                //printf("* ");
                if (/*count < SMEM_DIFF_BLK_MAX &&*/ g_bufMatrix[addrStartDest + index] != MATRIX_TRUE)
                {
                    //printf(". ");
                    if (g_bufCount[index] < min)
                    {
                        min = g_bufCount[index];
                        blk2Send = index;
                    }
                    count++;
                }
            }
            else
            {
                iAmDone = FALSE;
            }
        }

        //increase global done count
        if (iAmDone == TRUE) atomicAdd(g_doneCount, 1);

        #ifdef DEBUG_KERNEL
        //If there is something to transmit, then transmit a random block.
        printf("##start##tid %d: count %d, receiver=%d\n", tid, count, receiver);
        #endif
        
        
        if (count > 0)
        {
            //Transmit the block to receiver, any conflict here?
            //#ifdef DEBUG_KERNEL
            //printf("tid %d: receiverPeer=%d, blk2Send=%d\n", tid, receiver, blk2Send);
            //#endif
            //g_bufMatrix[addrStartDest + blk2Send] = MATRIX_TRUE;
            //update g_bufCount in the neighborhood
            g_bufUpdate[addrStartDest + blk2Send] = 1; 
        }
        //printf(" ==end==\n");
    }
}

#endif

#endif /* __GRF_KERNEL__ */
