#ifndef  __RUB_KERNEL__
#define __RUB_KERNEL__

#ifndef _D_INTRND_
#define _D_INTRND_
inline __device__ UINT intrnd (UINT seed) // 1<=seed<m
{
    UINT const a = 16807; //ie 7**5
    //UINT const m = 16777215; //ie 2**24-1
    //seed = (long(seed * a)) % m;
    seed = (UINT) __umul24(seed, a);
    return seed;
}
#endif /*_D_INTRND_*/

/*================== Shared memory based design =============*/
#ifdef RUB_SHARED_MEM
/* RUB algorithm with share memory used for optimization */
__global__ void d_p2pRUB_smem(USHORT *g_neighList, BOOL *g_bufMatrix, UINT *g_doneCount, UINT seed)
{
    __shared__ USHORT s_blkDiffList[BLK_SZ][SMEM_DIFF_BLK_MAX];
    USHORT              receivePeer;
    UINT              blkIndex2Send;
    UINT              index, count, tmp;
    UINT              addrStartSrc, addrStartDest;
    UINT              iAmDone  = FALSE;
    UINT              tid      = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;

    if(tid < NUM_PEERS)
    {
        // Select a random target peer for transmission
#ifdef RANDOM_SELECTION
        index         = __umul24(tid, NUM_PEERS); // just reuse index to save registers
        tmp           = intrnd(seed+tid) % g_neighList[index]; 
        receivePeer   = g_neighList[index + tmp + 1];
#else
        receivePeer   = (tid+1) % NUM_PEERS;
#endif
    
        count         = 0;
        iAmDone       = TRUE;
        addrStartSrc  = __umul24(tid, NUM_BLOCKS);
        addrStartDest = __umul24(receivePeer, NUM_BLOCKS);
        
        //Compare and list out the different buffer blocks
        for(index = 0; index < NUM_BLOCKS && (iAmDone == TRUE || count < SMEM_DIFF_BLK_MAX); index++)
        {
            //printf("start tid %d: target %d, index %d, src %d, target %d \n", tid, receivePeer, index, s_blkListSrc[threadIdx.x][index], g_bufMatrix[receivePeer*NUM_BLOCKS+index]);
            if (g_bufMatrix[addrStartSrc + index] == MATRIX_TRUE)
            {
                 if(count < SMEM_DIFF_BLK_MAX && g_bufMatrix[addrStartDest + index] == MATRIX_FALSE)
                {
                    s_blkDiffList[threadIdx.x][count] = index;
                    count++;
                }
            }
            else
            {
                iAmDone = FALSE;
            }
        }

        //increase global done count
        if (iAmDone == TRUE) atomicAdd(g_doneCount, 1);
    
        //__syncthreads();
        #ifdef DEBUG_KERNEL
        //If there is something to transmit, then transmit a random block.
        printf("##start##tid %d: count %d, receivePeer=%d\n", tid, count, receivePeer);
        #endif
        if (count > 0)
        {
#ifdef RANDOM_SELECTION
            //blkIndex2Send = s_blkDiffList[threadIdx.x][rand() % count];
            index = intrnd(seed+tid+1) % count;
#else
            index = 0;
#endif

            blkIndex2Send = s_blkDiffList[threadIdx.x][index];

            //Transmit the block to receiver, any conflict here?
            #ifdef DEBUG_KERNEL
            printf("==end==tid %d: receiverPeer=%d, blkIndex2Send=%d\n", tid, receivePeer, blkIndex2Send);
            #endif
            //g_bufMatrix[addrStartDest + blkIndex2Send] = g_bufMatrix[addrStartSrc + blkIndex2Send];
            g_bufMatrix[addrStartDest + blkIndex2Send] = MATRIX_TRUE;
        }
    
        //__syncthreads();
    }
}

#endif

#ifdef RUB_NO_SHARED_MEM
/*===================== No shared memory based design =============*/
/* RUB algorithm without using share memory, but scan twice. */
__global__ void d_p2pRUB_no_smem(USHORT *g_neighList, BOOL *g_bufMatrix, UINT *g_doneCount, UINT seed)
{
    //__shared__ USHORT s_blkDiffList[BLK_SZ][SMEM_DIFF_BLK_MAX];
    USHORT            receivePeer;
    UINT              blkIndex2Send;
    UINT              index, count, tmp;
    UINT              addrStartSrc, addrStartDest;
    UINT              iAmDone  = FALSE;
    UINT              tid      = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;

    if(tid < NUM_PEERS) // one thread for each peer
    {
        // Select a random target peer for transmission
        index         = __umul24(tid, NUM_PEERS); // just reuse index to save registers
        tmp           = intrnd(seed+tid) % g_neighList[index]; 
        receivePeer   = g_neighList[index + tmp + 1];
   
        count         = 0;
        iAmDone       = TRUE;
        addrStartSrc  = __umul24(tid, NUM_BLOCKS);
        addrStartDest = __umul24(receivePeer, NUM_BLOCKS);
 
        //Compare and list out the different buffer blocks
        for(index = 0; index < NUM_BLOCKS; index++)
        {
            //printf("start tid %d: target %d, index %d, src %d, target %d \n", tid, receivePeer, index, s_blkListSrc[threadIdx.x][index], g_bufMatrix[receivePeer*NUM_BLOCKS+index]);
            if (g_bufMatrix[addrStartSrc + index] == MATRIX_TRUE)
            {
                if(g_bufMatrix[addrStartDest + index] == MATRIX_FALSE)
                {
                    //s_blkDiffList[threadIdx.x][count] = index;
                    count++;
                }
            }
            else
            {
                iAmDone = FALSE;
            }
        }
        
        //increase global done count
        if (iAmDone == TRUE) atomicAdd(g_doneCount, 1);
    
        
        #ifdef DEBUG_KERNEL
        //If there is something to transmit, then transmit a random block.
        printf("##start##tid %d: count %d, receivePeer=%d\n", tid, count, receivePeer);
        #endif
        if (count > 0)
        {
            UINT count1 = intrnd(seed+tid+1) % count + 1; // a random number in [1, count]
            count = 0;
            for(index = 0; index < NUM_BLOCKS && count < count1; index++)
            {
                //printf("start tid %d: target %d, index %d, src %d, target %d \n", tid, receivePeer, index, s_blkListSrc[threadIdx.x][index], g_bufMatrix[receivePeer*NUM_BLOCKS+index]);
                if (g_bufMatrix[addrStartSrc + index] == MATRIX_TRUE
                    && g_bufMatrix[addrStartDest + index] == MATRIX_FALSE)
                {
                    count++;
                }
            }
            blkIndex2Send = index - 1;

            //Transmit the block to receiver, any conflict here?
            #ifdef DEBUG_KERNEL
            printf("==end==tid %d: receiverPeer=%d, blkIndex2Send=%d\n", tid, receivePeer, blkIndex2Send);
            #endif
            g_bufMatrix[addrStartDest + blkIndex2Send] = MATRIX_TRUE;
        }
    }
}
#endif

/*======================== Buffer Per Thread design ===================*/
/*========================  Some bugs need to fix   ===================*/
#ifdef RUB_BUFFER_PER_THREAD

__global__ void d_p2pRUB_bufPerThread(USHORT *g_neighList, BOOL *g_bufMatrix, UINT *g_doneCount, UINT seed)
{
    //__shared__ USHORT s_blkDiffList[BLK_SZ][SMEM_DIFF_BLK_MAX];
    __shared__ USHORT receivePeer, sendPeer;
    __shared__ UINT   blk2Send;
    __shared__ UINT   /*addrStartSrc,*/ addrStartDest;
    __shared__ UINT   iAmDone;

    UINT              blkIndex;
    UINT              tid      = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;

    if (tid < NUM_PEERS * NUM_BLOCKS)
    {
        blkIndex = tid % NUM_BLOCKS;
        
        // Peer Operations
        if (blkIndex == 0) // NUM_BLOCKS == BLK_SZ
        {
            blk2Send = NUM_BLOCKS;
            iAmDone = TRUE;
            sendPeer      = tid / NUM_BLOCKS; 
            // Select a random target peer for transmission
            UINT tmp      = intrnd(seed+sendPeer) % g_neighList[sendPeer * NUM_PEERS]; 
            //UINT tmp = 0;
            receivePeer   = g_neighList[sendPeer*NUM_PEERS + tmp + 1];
            //addrStartSrc  = tid;
            addrStartDest = __umul24(receivePeer, NUM_BLOCKS);
            //printf("tid : %d, sendPeer %d, receivePeer %d\n", tid, sendPeer, receivePeer);
        }
        
        // Block Operations: buffer comparison      
        __syncthreads();
        if (g_bufMatrix[tid] == MATRIX_TRUE)
        {
            if(g_bufMatrix[addrStartDest + blkIndex] == MATRIX_FALSE)
            {
                atomicExch(&blk2Send, blkIndex);
            }
        }
        else
        {
            iAmDone = FALSE;
        }    
        __syncthreads();
        
        //Peer operations
        if (blkIndex == 1)
        {
            if (blk2Send < NUM_BLOCKS)
            {     
                //printf("tid : %d, Peer %d sends a block to Peer %d\n", tid, sendPeer, receivePeer);
                g_bufMatrix[addrStartDest + blk2Send] = MATRIX_TRUE;
            }
        }
        if (blkIndex == 2)
        {
            //increase global done count
            if (iAmDone == TRUE) atomicAdd(g_doneCount, 1);
        }
    }
}
#endif

#endif /* __RUB_KERNEL__ */
