#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>
#include <cutil.h>
#include "cutil_math.h"
//#include <cuda_runtime.h>


#define   RANDOM_SELECTION // used for randomly select the peer and blocks
//#define DEBUG_KERNEL
//#define DEBUG_CPU

#define     TRUE               1
#define     FALSE              0
#define     MATRIX_TRUE        1
#define     MATRIX_FALSE       0
#define     SHARED_MEM_SZ      (1<<14)   // 16KB

//=========== adjustable settings ==============
//#define     GRF_PEER_PER_THREAD
#define     GRF_DATABLOCK_PER_THREAD

#ifdef GRF_PEER_PER_THREAD
  #define   BLK_SZ             512
#else
  #define   BLK_SZ             512         
#endif

#define     NUM_PEERS          1024 //(1<<14)  // # of peers
#define     NUM_BLOCKS         512  // # of buffer blocks, must be < 65535

#define     MAX(a,b)           (((a) >= (b))? (a):(b))
#define     MIN(a,b)           (((a) <= (b))? (a):(b))
#define     iDivUp(a, b)       (((a)+(b)-1)/(b))

// Max number of different blocks to save on shared memory
#define     SMEM_DIFF_BLK_MAX  NUM_BLOCKS        // 98 for 64 blks, 226 for 32 blks
//#define     GRID_SZ            iDivUp(NUM_PEERS, BLK_SZ)

#define     GRAPH_PROB         0.5      // Topology: Random Erdos-Renyi Graph with
                                        // parameter p, p=1 means complete graph.
#define     MAX_ROUNDS         (1<<16)
#define     MIN_NEIGHBOUR      3

//==============================================
#if (NUM_BLOCKS > 0xFFFF) // max of USHORT data type
#error "ERROR: NUM_BLOCKS exceeded the limitation"
#endif

//#if ((SMEM_DIFF_BLK_MAX * BLK_SZ << 1) > SHARED_MEM_SZ) // max of USHORT data type
//#error "ERROR: NUM_BLOCKS exceeded the limitation"
//#endif

typedef unsigned int        UINT;
typedef unsigned short      USHORT;
typedef unsigned char       UCHAR;
typedef unsigned char       BOOL;

//#include "RUB_kernel.cu"
//#include "LRF_kernel.cu"
#include "GRF_kernel.cu"

//============= Functions ======================
void myFree(BOOL **ptr, UINT num)
{
    UINT index;
    if (ptr)
    {
        for(index = 0; index < num; index++)
        {
            if (ptr[index])
                free(ptr[index]);
        }
        free(ptr);
    }
}

void myFree(UINT **ptr, UINT num)
{
    UINT index;
    if (ptr)
    {
        for(index = 0; index < num; index++)
        {
            if (ptr[index])
                free(ptr[index]);
        }
        free(ptr);
    }
}

void myFree(BOOL *ptr)
{
    if (ptr)
        free(ptr);
}

void myFree(USHORT *ptr)
{
    if (ptr)
        free(ptr);
}

void myFree(UINT *ptr)
{
    if (ptr)
        free(ptr);
}

void * myMalloc(size_t size)
{
    void *ptr = malloc(size);
    if (ptr == NULL)
    {
        printf("ERROR: failed to allocated memory\n");
        return(NULL);
    }

    memset(ptr, 0, size);
    return(ptr);
}

UINT myRand(UINT range)
{
    UINT value = rand();
    return(value % range);
}

//Returns a pseudo-random number in (0, 1).
float rand01(void)
{
    float z;
    
    UINT mm = (1 << 30) - 1;
    UINT aa = 16807;
    UINT qq = mm / aa;
    UINT rr = mm % aa;
    UINT seed = rand();
    
    seed = aa * (seed % qq) - rr * (seed / qq);
    if (seed <= 0)
        seed = seed + mm;

    z = (seed * 1.0) / mm;
    if (z > 1.0)
        z -= floor(z);

    return z;
}

void printMatrix(UINT **matrix, UINT num_rows, UINT num_cols)
{
    UINT i, j;
    for(i = 0; i < num_rows; i++)
    {    
        for(j = 0; j < num_cols; j++)
            printf("%d ", matrix[i][j]);

        printf("\n");
    }
    printf("\n");    
}

UINT verifyBufferMatrix(BOOL *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    for(i = 0; i < num_rows * num_cols; i++)
    {    
        if (matrix[i] != MATRIX_TRUE)
    {
            printf("WARNING: matrix contain non-finished buffers, row=%d, blk=%d\n", 
                i / num_cols, i % num_cols);
        return(FALSE);
    }
    }

    return(TRUE);
}

void printMatrix(BOOL *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    for(i = 0; i < num_rows * num_cols; i++)
    {    
        if (matrix[i] == MATRIX_TRUE)
                printf("1 ");
        else
            printf("0 ");
    
            if ((i+1) % num_cols == 0)
                printf("\n");
    }
    printf("\n");    
}

void printMatrix(UINT *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    UINT *ptr = matrix;
    printf("ROW 0:    ");
    for(i = 0; i < num_rows * num_cols; i++, ptr++)
    {    
        printf("%d ", *ptr);

        if ((i+1) % num_cols == 0)
            printf("\nROW %d:    ", (i+1)/num_cols);
    }
    printf("\n");
}

void printMatrix(USHORT *matrix, UINT num_rows, UINT num_cols)
{
    UINT i;
    USHORT *ptr = matrix;
    printf("ROW 0:    ");
    for(i = 0; i < num_rows * num_cols; i++, ptr++)
    {    
        printf("%d ", *ptr);

        if ((i+1) % num_cols == 0)
            printf("\nROW %d:    ", (i+1)/num_cols);
    }
    printf("\n");
}

//Generate a random adjacency matrix
void randAdjacency(UINT numPeers, float prob, UINT **peerConnTbl)
{
    UINT  i,j;

    for(i=0; i<numPeers; i++)
    {    
        for(j=i; j<numPeers; j++)
        {
            if( (prob == 1.0 || rand01() < prob) && (i != j) )
            {
                peerConnTbl[i][j] = 1;
                peerConnTbl[j][i] = 1;
            }
            else
            {
                peerConnTbl[i][j] = 0;
                peerConnTbl[j][i] = 0;
            }            
        }
    }
}

//Random/Complete Graph Generator
//Peering information are in peerConnTbl and neighTbl 
//Returns the minimum number of neighbors a peer has.
int randTopoGen(UINT numPeers, float prob, UINT **peerConnTbl, USHORT *neighTbl)
{
    USHORT           i, j;
    USHORT             minNumNeighbor = numPeers - 1;

    //Generate a random adjacency matrix
    randAdjacency(numPeers, prob, peerConnTbl);
    //Generate the Neighbor Information matrix [NeighTbl(i,j)] based on peerConnTbl
    for(i = 0; i < numPeers; i++)
    {
        USHORT l = 1;

        memset(&neighTbl[i*numPeers], 0, sizeof(USHORT) * numPeers);
        for(j = 0; j < numPeers; j++)
        {
            if(peerConnTbl[i][j] == TRUE)
            {
                neighTbl[i*numPeers+l] = j;
                l++;
            }
        }

        neighTbl[i*numPeers] = l-1;
        minNumNeighbor = MIN(minNumNeighbor, neighTbl[i*numPeers]);
    }
    return minNumNeighbor;
}

//Find what is in buffSource, but is not in buffTarget. 
//diff stores the positions (block_ID=i) in the buffermap that buffSource[i]==1 && buffTarget[i]==0
//returns the length of diff.
UINT buffDiff(UINT* diff, BOOL* buffSource, BOOL* buffTarget, UINT length)
{
    UINT i, count=0;
    for(i = 0; i < length; i++)
    {
        if(buffSource[i] == MATRIX_TRUE && buffTarget[i] != MATRIX_TRUE)
        {
            diff[count]=i;
            count++;
        }
    }
    return count;
}

//Initialize the buffer maps. N is the number of peers, k is number of blocks to
//be distributed, B[i] is the pointer to Peer i's buffer map.
//The initial buffer map at each peer (B(i, 1:1:k) is the buffer map of Peer i)
void buffInitialize(UINT numPeers, UINT numBlocks, BOOL *h_bufferMatrix)
{ 
    memset(&h_bufferMatrix[0], MATRIX_TRUE, NUM_BLOCKS * sizeof(BOOL));
    memset(&h_bufferMatrix[NUM_BLOCKS], MATRIX_FALSE, (numPeers-1) * NUM_BLOCKS * sizeof(BOOL));
}

void buffInitialize(UINT numPeers, UINT numBlocks, BOOL *h_bufferMatrix, BOOL *h_bufUpdate)
{ 
    memset(&h_bufferMatrix[0], MATRIX_TRUE, NUM_BLOCKS * sizeof(BOOL));
    memset(&h_bufferMatrix[NUM_BLOCKS], MATRIX_FALSE, (numPeers-1) * NUM_BLOCKS * sizeof(BOOL));
    memset(&h_bufUpdate[0], 0, numPeers * NUM_BLOCKS * sizeof(BOOL));    
}

void bufCntInit(UINT numBlocks, UINT *bufCount)
{
    UINT l;
    // Initialize bufCount
    for (l = 0; l < numBlocks; l++)
        bufCount[l] = 1;
}

// Copy matrix
void copyMatrix(BOOL **src, BOOL **dest, UINT numRows, UINT numCols)
{
    UINT i, j;
    for (i = 0; i < numRows; i++)
        for (j = 0; j < numCols; j++)
            dest[i][j] = src[i][j];
}

//Protocol 1 Random Useful Block (RUB): In each round, each peer first 
//chooses a random neighbor, and then transmits a random useful block to the receiver.
//If preSetSchedu is 1, use the pre-generated source and target schedules given 
//in int** source and int** target;
//otherwise, generate the source and target schedules on the go and store them
//in int** source and int** target.
UINT p2pRUB(UINT numPeers, UINT numBlocks, BOOL* bufferMatrix, USHORT* h_neighbourList)
{
    UINT   r, i, j;
    USHORT   transmitter, receiver;
    USHORT randNeighIndex;
    UINT   *Set4Trans = NULL;
    UINT   lengthDiff;
    BOOL   *bufMatrixRecv;
    BOOL   finish = FALSE;
    
    // Allocate memory for bufMatrixRecv
    bufMatrixRecv = (BOOL *) myMalloc(numPeers * numBlocks * sizeof(BOOL));
    if (bufMatrixRecv == NULL)
        return(0);

    // Initialize bufMatrixRecv
    memcpy(bufMatrixRecv, bufferMatrix, sizeof(BOOL) * numPeers * numBlocks);
    
    for (r = 0; finish == FALSE && r < MAX_ROUNDS; r++)
    {
        //Round r
        //printf("Round %d\n", r);
        for (i = 0; i < numPeers; i++)
        {            
            transmitter = i;

            // assume every peer will perform in Round r
            //randNeighIndex = (UINT) floor(rand01() * h_neighbourList[i][0]);
            randNeighIndex = myRand(h_neighbourList[i*numPeers]);
            receiver = (UINT) h_neighbourList[i*numPeers+randNeighIndex+1];

      //printf("round %d, peer %d, totalNeighbours %d, randNeighIndex=%d, target=%d\n", r, i, h_neighbourList[i][0], randNeighIndex, receiver;
            
            //Find the blocks that transmitter has and receiver doesn't have. 
            //Store them in Set4Trans, lengthDiff is the length of Set4Trans
            Set4Trans = (UINT *) myMalloc(sizeof(UINT) * numBlocks);
            if (Set4Trans == NULL)
            {
                myFree(bufMatrixRecv);
                return(0);
            }

            lengthDiff = buffDiff(Set4Trans, &bufferMatrix[transmitter*numBlocks], &bufMatrixRecv[receiver*numBlocks], numBlocks);
            
            //If there is something to trannsmit, then transmit a random block.
            if (lengthDiff > 0)
            {
                //UINT blockID = Set4Trans[(UINT)(rand01() * lengthDiff)];
                UINT blockID = Set4Trans[myRand(lengthDiff)];
                
                //Transmit the block with ID blockID from transmitter to receiver by updating bufMatrixRecv
                //printf("Round %d: Peer %d transmits block %d to Peer %d\n", r, transmitter, blockID, receiver);
                //printf(". ");
                bufMatrixRecv[receiver*numBlocks+blockID] = bufferMatrix[transmitter*numBlocks+blockID];
            }
            free(Set4Trans);
        }

        //update bufferMatrix
        //copyMatrix(bufMatrixRecv, bufferMatrix, numPeers, numBlocks);
        memcpy(bufferMatrix, bufMatrixRecv, sizeof(BOOL) * numPeers * numBlocks);
        
        //If all the peer buffers are filled, then break.
        finish = TRUE;
        for (j = 0; j < numPeers * numBlocks && finish == TRUE; j++)
            if (bufferMatrix[j] != MATRIX_TRUE)
                finish = FALSE;
        
        //printf("\n");
    } // end of Round r

    myFree(bufMatrixRecv);

    return r;
}



//Protocol 2 Local Rarest First (LRF): In each round, each peer first 
//chooses a random neighbor, and then transmits a random useful block to the receiver that is the rarest
//within the neighbohood of the transmitter.
UINT p2pLRF(UINT numPeers, UINT numBlocks, BOOL* bufferMatrix, USHORT* h_neighbourList)
{
    UINT     r, i, j, l;
    USHORT   transmitter, receiver;
    USHORT   randNeighIndex;
    UINT     lengthDiff;
    BOOL     *bufMatrixRecv;
    BOOL     iAmDone;
    UINT     numFinish = 0;
    USHORT   *bufCount;
    UINT     min = numPeers + 1;
    UINT     blk2Send;
        
    // Allocate memory for bufMatrixRecv
    bufMatrixRecv = (BOOL *) myMalloc(numPeers * numBlocks * sizeof(BOOL));
    if (bufMatrixRecv == NULL)
        return(0);

    // Initialize bufMatrixRecv
    memcpy(bufMatrixRecv, bufferMatrix, sizeof(BOOL) * numPeers * numBlocks);
    
    // Allocate memory for bufCount
    bufCount = (USHORT *) myMalloc(numPeers * numBlocks * sizeof(USHORT));
    if (bufCount == NULL)
        return(0);

    // Initialize bufCount
    for (l = 0; l < numBlocks; l++)
        bufCount[l] = 1;
    
    for (j = 1; j < numPeers; j++)
        for (l = 0; l < numBlocks; l++)
            bufCount[j*numBlocks + l] = 0;
    
    for (i = 0; i < h_neighbourList[0]; i++)
    {
        j = h_neighbourList[i+1];
        for (l = 0; l < numBlocks; l++)
            bufCount[j*numBlocks + l] = 1;
    }


    for (r = 0; numFinish < numPeers && r < MAX_ROUNDS; r++)
    {
        //Round r
        //printf("Round %d\n", r);
        numFinish = 0;
        
        for (i = 0; i < numPeers; i++)
        {            
            transmitter = i;

            randNeighIndex = myRand(h_neighbourList[i*numPeers]);
            receiver = (UINT) h_neighbourList[i*numPeers+randNeighIndex+1];

      //printf("round %d, peer %d, totalNeighbours %d, randNeighIndex=%d, target=%d\n", r, i, h_neighbourList[i][0], randNeighIndex, receiver;
            lengthDiff = 0;
            iAmDone = TRUE;
            min = numPeers + 1;
                            
            for (j = 0; j < numBlocks; j++)
            {
                if (bufferMatrix[transmitter*numBlocks + j] == MATRIX_TRUE)
                {
                    if (bufMatrixRecv[receiver*numBlocks + j] != MATRIX_TRUE)
                    {
                        if (bufCount[transmitter*numBlocks + j] < min)
                        {
                            min = bufCount[transmitter*numBlocks + j];
                            blk2Send = j;
                        }
                        lengthDiff ++;
                    }
                }
                else
                {
                    iAmDone = FALSE;
                }
            }
            
            if (iAmDone == TRUE) numFinish ++;
            
            //If there is something to trannsmit, then transmit blk2Send.
            if (lengthDiff > 0)
            {
                //printf("Round %d: Peer %d transmits block %d to Peer %d\n", r, transmitter, blk2Send, receiver);
                //printf(". ");
                bufMatrixRecv[receiver*numBlocks+blk2Send] = bufferMatrix[transmitter*numBlocks+blk2Send];
                //update bufCount
                bufCount[receiver*numBlocks+blk2Send] ++;
                for (l = 0; l < h_neighbourList[receiver*numPeers]; l++)
                {
                    j = h_neighbourList[receiver*numPeers+l+1];
                    bufCount[j*numBlocks + blk2Send] ++;
                } 
            }
        }

        //update bufferMatrix
        memcpy(bufferMatrix, bufMatrixRecv, sizeof(BOOL) * numPeers * numBlocks);
        
       //printf("\n");
    } // end of Round r

    myFree(bufMatrixRecv);

    return r;
}




//Protocol 3 Global Rarest First (GRF): In each round, each peer first 
//chooses a random neighbor, and then transmits a random useful block to the receiver that is the rarest
//from the perspective of the entire network.
UINT p2pGRF(UINT numPeers, UINT numBlocks, BOOL* bufferMatrix, USHORT* h_neighbourList)
{
    UINT     r, i, j;
    USHORT   transmitter, receiver;
    USHORT   randNeighIndex;
    UINT     lengthDiff;
    BOOL     *bufMatrixRecv;
    BOOL     *bufUpdate;
    BOOL     iAmDone;
    UINT     numFinish = 0;
    USHORT   *bufCount;
    UINT     min = numPeers + 1;
    UINT     blk2Send;
        
    // Allocate memory for bufMatrixRecv and bufUpdate
    bufMatrixRecv = (BOOL *) myMalloc(numPeers * numBlocks * sizeof(BOOL));
    if (bufMatrixRecv == NULL)
        return(0);
    bufUpdate = (BOOL *) myMalloc(numPeers * numBlocks * sizeof(BOOL));
    if (bufUpdate == NULL)
        return(0);

    // Initialize bufMatrixRecv 
    memcpy(bufMatrixRecv, bufferMatrix, sizeof(BOOL) * numPeers * numBlocks);
    
    // Allocate memory for bufCount
    bufCount = (USHORT *) myMalloc(numBlocks * sizeof(USHORT));
    if (bufCount == NULL)
        return(0);

    // Initialize bufCount
    for (i = 0; i < numBlocks; i++)
        bufCount[i] = 1;
        
    for (r = 0; numFinish < numPeers && r < MAX_ROUNDS; r++)
    {
        //Round r
        //printf("Round %d\n", r);
        numFinish = 0;
        
        memset(bufUpdate, 0, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL));

        for (i = 0; i < numPeers; i++)
        {            
            transmitter = i;

            randNeighIndex = myRand(h_neighbourList[i*numPeers]);
            receiver = (UINT) h_neighbourList[i*numPeers+randNeighIndex+1];

      //printf("round %d, peer %d, totalNeighbours %d, randNeighIndex=%d, target=%d\n", r, i, h_neighbourList[i][0], randNeighIndex, receiver;
            lengthDiff = 0;
            iAmDone = TRUE;
            min = numPeers + 1;
                            
            for (j = 0; j < numBlocks; j++)
            {
                if (bufferMatrix[transmitter*numBlocks + j] == MATRIX_TRUE)
                {
                    if (bufferMatrix[receiver*numBlocks + j] != MATRIX_TRUE)
                    {
                        if (bufCount[j] < min)
                        {
                            min = bufCount[j];
                            blk2Send = j;
                        }
                        lengthDiff ++;
                    }
                }
                else
                {
                    iAmDone = FALSE;
                }
            }
            
            if (iAmDone == TRUE) 
            {
                numFinish ++;
                //printf("Transmitter %d, numFinish = %d\n", transmitter, numFinish);
            }
            
            //If there is something to trannsmit, then transmit blk2Send.
            if (lengthDiff > 0)
            {
                //printf("Round %d: Peer %d transmits block %d to Peer %d\n", r, transmitter, blk2Send, receiver);
                //printf(". ");
                bufMatrixRecv[receiver*numBlocks+blk2Send] = MATRIX_TRUE;
                bufUpdate[receiver*numBlocks+blk2Send] = 1;
                //printf("Block %d has %d copies \n", blk2Send, bufCount[blk2Send]);
            }
            
        }
        
        //update bufCount
        for (i = 0; i < numPeers; i++)
        {
            for (j = 0; j < numBlocks; j++)
            {
                /*if (bufMatrixRecv[i*numBlocks+j] == MATRIX_TRUE && bufferMatrix[i*numBlocks+j] == MATRIX_FALSE)
                {
                    bufCount[j]++;
                }*/
                bufCount[j] += bufUpdate[i*numBlocks+j];
            }    
        }
        

        //update bufferMatrix
        memcpy(bufferMatrix, bufMatrixRecv, sizeof(BOOL) * numPeers * numBlocks);
        
       //printf("\n");
    } // end of Round r

    myFree(bufMatrixRecv);

    return r;
}


/*-----------------------------------------------*/
/*              Main function                    */
/*-----------------------------------------------*/
int main( int argc, char** argv) 
{
    int               dev;
    UINT              timer        = 0;
    float             GPUtotaltime = 0;
    float             CPUtotaltime = 0;
    cudaError_t       d_ret;
    UINT              i;
    UINT              tmp1 = 0;
    UINT              tmp2 = 0;
    UINT              minNumNeighbor;
    UINT              cpuDelayGRF, gpuDelayGRF;
    UINT              **peerConnTable  = NULL;
    USHORT            *h_neighbourList  = NULL;
    BOOL              *h_bufferMatrix   = NULL;
    BOOL              *h_bufUpdate = NULL;
    UINT              *h_bufCount = NULL;
    BOOL              *d_bufferMatrix   = NULL;
    BOOL              *d_bufUpdate = NULL;
    UINT              *d_bufCount = NULL;
    USHORT            *d_neighbourList  = NULL;
    UINT              *d_peerDoneCount;
    UINT              *d_roundCount;
    dim3              blk_Peer, grid_Peer, blk_Data, grid_Data;
    

    srand(time(NULL));

    // start timer for CPU calculation
    CUT_DEVICE_INIT(argc, argv);

    /* check GPU device */
    d_ret      = cudaGetDevice(&dev);
    if (d_ret != cudaSuccess)
    {
        printf("CUDA: get device failed\n");
        exit(10);
    }


    printf("Allocating CPU memory");
    // Allocate memory for topology matrices and Neighbor
    peerConnTable  = (UINT **) myMalloc(NUM_PEERS * sizeof(UINT *));

    if (peerConnTable == NULL)
        exit(11);
    
    printf(" . . .");
    for(i=0; i<NUM_PEERS; i++)
    {
        peerConnTable[i] = (UINT *) myMalloc(NUM_PEERS * sizeof(UINT));
        if (peerConnTable[i] == NULL)
        {
            myFree(peerConnTable, i);
            exit(12);
        }
    }

    printf(" . . .\n");
    h_neighbourList  = (USHORT *) myMalloc(NUM_PEERS * NUM_PEERS * sizeof(USHORT));
    if (h_neighbourList == NULL)
    {
        myFree(peerConnTable, NUM_PEERS);
        exit(13);
    }
    
    printf("Genrating random graph . . . ");
    // Generate an Erdos-Renyi random graph with N nodes and parameter p
    do
    {
        minNumNeighbor = randTopoGen(NUM_PEERS, GRAPH_PROB, peerConnTable, h_neighbourList);
    } while (minNumNeighbor <= MIN_NEIGHBOUR);

    printf(" min neighbors: %d\n", minNumNeighbor);
    
    // Display the generated topology
    //printMatrix(peerConnTable, NUM_PEERS, NUM_PEERS);
    //printMatrix(h_neighbourList, NUM_PEERS, NUM_PEERS);

    myFree(peerConnTable, NUM_PEERS);

    printf("Allocating memory for buffer matrix . . .\n");
    // Allocate memory for the block matrix, which stores peer buffer states
    // each bit represents one buffer block to save memory usage
    h_bufferMatrix = (BOOL *) myMalloc(NUM_PEERS * NUM_BLOCKS * sizeof(BOOL));
    h_bufUpdate = (BOOL *) myMalloc(NUM_PEERS * NUM_BLOCKS * sizeof(BOOL));
    h_bufCount = (UINT *) myMalloc(NUM_BLOCKS * sizeof(UINT));
    if (h_bufferMatrix == NULL || h_bufCount == NULL || h_bufUpdate == NULL)
    {
        myFree(h_neighbourList);
        exit(14);
    }

    // Initialize buffer maps
    buffInitialize(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix, h_bufUpdate);
    bufCntInit(NUM_BLOCKS, h_bufCount);
    //printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    
    printf("Allocating memory on device");

    // Allocate memory on Device and copy data over
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_neighbourList, NUM_PEERS * NUM_PEERS * sizeof(USHORT)));
    if (d_neighbourList == NULL)
    {
        myFree(h_neighbourList);
        myFree(h_bufferMatrix);
        exit(15);
    }
    
    printf(" . . .");
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_bufferMatrix, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL)));
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_bufUpdate, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL)));
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_bufCount, NUM_BLOCKS * sizeof(UINT)));
    if (d_bufferMatrix == NULL || d_bufCount == NULL || d_bufUpdate == NULL)
    {
        myFree(h_neighbourList);
        myFree(h_bufferMatrix);
        myFree(h_bufCount);
        CUDA_SAFE_CALL(cudaFree(d_neighbourList));
        exit(16);
    }

    printf(" . . .\n");
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_peerDoneCount, sizeof(UINT)));
    CUDA_SAFE_CALL(cudaMalloc((void **) &d_roundCount, sizeof(UINT)));

    if (d_peerDoneCount == NULL || d_roundCount == NULL )
    {
        myFree(h_neighbourList);
        myFree(h_bufferMatrix);
        CUDA_SAFE_CALL(cudaFree(d_neighbourList));
        CUDA_SAFE_CALL(cudaFree(d_bufferMatrix));
        exit(18);
    }

    printf("copying data to device . . .");

    CUDA_SAFE_CALL(cudaMemcpy(d_peerDoneCount, &tmp1, sizeof(UINT), cudaMemcpyHostToDevice ));
    CUDA_SAFE_CALL(cudaMemcpy(d_roundCount, &tmp2, sizeof(UINT), cudaMemcpyHostToDevice ));

    printf(" . . .\n");
    CUDA_SAFE_CALL(cudaMemcpy(d_bufferMatrix, h_bufferMatrix, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL), cudaMemcpyHostToDevice ));
    CUDA_SAFE_CALL(cudaMemcpy(d_bufUpdate, h_bufUpdate, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL), cudaMemcpyHostToDevice ));
    CUDA_SAFE_CALL(cudaMemcpy(d_bufCount, h_bufCount, NUM_BLOCKS * sizeof(UINT), cudaMemcpyHostToDevice ));
    CUDA_SAFE_CALL(cudaMemcpy(d_neighbourList, h_neighbourList, NUM_PEERS * NUM_PEERS * sizeof(USHORT), cudaMemcpyHostToDevice ));

 #ifdef DEBUG_KERNEL
    printf("buffer matrix source on device\n");
    printMatrix(d_bufferMatrix, NUM_PEERS, NUM_BLOCKS);
 #endif

    
    CUT_SAFE_CALL( cutCreateTimer( &timer));
       
    printf("============== GRF on GPU & CPU  =============\n");


    USHORT blockSZ = BLK_SZ;
#ifdef PEER_PER_THREAD
    USHORT gridSZ = iDivUp(NUM_PEERS, blockSZ);;
    blk_Peer  = dim3(blockSZ);
    grid_Peer = dim3(gridSZ);

    USHORT blockSZ_adv = 512;
    USHORT gridSZ_adv = iDivUp(NUM_PEERS * NUM_BLOCKS, blockSZ_adv);
    blk_Data = dim3(blockSZ_adv);
    grid_Data = dim3(gridSZ_adv);

    printf("peer number %d, buffer blocks %d\n", NUM_PEERS, NUM_BLOCKS);
    printf("block size %d, grid size %d\n", blockSZ, gridSZ);
#else
    USHORT gridSZ = NUM_PEERS;
    blk_Peer = dim3(blockSZ);
    grid_Peer = dim3(gridSZ);
    
    blk_Data = dim3(BLK_SZ);
    grid_Data = dim3(iDivUp(NUM_PEERS * NUM_BLOCKS, BLK_SZ));
#endif

    printf("peer number %d, buffer blocks %d\n", NUM_PEERS, NUM_BLOCKS);
    printf("block size %d, grid size %d\n", blockSZ, gridSZ);

    cutResetTimer(timer);
    cutStartTimer(timer);
    for(gpuDelayGRF = 0; gpuDelayGRF < MAX_ROUNDS; gpuDelayGRF++)
    {
        #ifdef DEBUG_KERNEL
        printf("============= GPU round %d ==============\n", gpuDelayGRF);
        #endif
        
        d_p2pGRF <<<grid_Peer, blk_Peer>>> (d_neighbourList, d_bufferMatrix, d_bufCount, d_bufUpdate, d_peerDoneCount, gpuDelayGRF);
        CUDA_SAFE_CALL(cudaThreadSynchronize());
        
        // check if we have finished transmiting
        CUDA_SAFE_CALL(cudaMemcpy(&tmp1, d_peerDoneCount, sizeof(UINT), cudaMemcpyDeviceToHost ));
        //printf("Round %d: number of finished peers = %d\n", gpuDelayGRF, tmp1);
        if (tmp1 < NUM_PEERS)
        {
            //updating peer buffers and d_bufCount
            d_Update <<<grid_Data, blk_Data>>> (d_bufUpdate, d_bufferMatrix, d_bufCount);
            CUDA_SAFE_CALL(cudaThreadSynchronize());
            //printMatrix(d_bufferMatrix, NUM_PEERS, NUM_BLOCKS);
            //printf("\nbufCount:\n");
            //printMatrix(d_bufCount, 1, NUM_BLOCKS);
            tmp1 = 0;
            CUDA_SAFE_CALL(cudaMemcpy(d_peerDoneCount, &tmp1, sizeof(UINT), cudaMemcpyHostToDevice ));
        }
        else
            break;
    }
    cutStopTimer(timer);
    GPUtotaltime = cutGetTimerValue(timer);


    CUDA_SAFE_CALL(cudaMemcpy(h_bufferMatrix, d_bufferMatrix, NUM_PEERS * NUM_BLOCKS * sizeof(BOOL), cudaMemcpyDeviceToHost ));
#ifdef DEBUG_CPU
    printf("====== buffer matrix after kernel done =======\n");
    printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);
#endif
    verifyBufferMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    printf("GPU: GRF Rounds = %d, time = %.6f s\n", gpuDelayGRF, GPUtotaltime/1000);
    printf("GPU Processing Speed: %.6f s/round\n", GPUtotaltime/1000/gpuDelayGRF);

    CUDA_SAFE_CALL(cudaFree(d_neighbourList));
    CUDA_SAFE_CALL(cudaFree(d_bufferMatrix));
    CUDA_SAFE_CALL(cudaFree(d_bufUpdate));
    CUDA_SAFE_CALL(cudaFree(d_bufCount));
    CUDA_SAFE_CALL(cudaFree(d_peerDoneCount));
    //CUDA_SAFE_CALL(cudaFree(d_roundCount));

    /*
    //printf("============== RUB on CPU =============\n");
    // Initialize buffer maps
    buffInitialize(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix);
    //printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    // create timer
    cutResetTimer(timer);
    cutStartTimer(timer);

    // Disseminating blocks using RUB
    cpuDelayRUB = p2pRUB(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix, h_neighbourList);

    // stop timer
    cutStopTimer(timer);
    CPUtotaltime = cutGetTimerValue(timer);

    //printMatrix(B, N, k);
    printf("CPU: RUB Rounds = %d, time = %.6f s\n", cpuDelayRUB, CPUtotaltime/1000);
    

    //printf("============== LRF on CPU =============\n");
    // Initialize buffer maps
    buffInitialize(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix);
    //printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    // create timer
    cutResetTimer(timer);
    cutStartTimer(timer);

    // Disseminating blocks using LRF
    cpuDelayLRF = p2pLRF(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix, h_neighbourList);

    // stop timer
    cutStopTimer(timer);
    CPUtotaltime = cutGetTimerValue(timer);

    printf("CPU: LRF Rounds = %d, time = %.6f s\n", cpuDelayLRF, CPUtotaltime/1000);
       printf("GPU speedup: %.2f\n", CPUtotaltime/GPUtotaltime);
    */

    //printf("============== GRF on CPU =============\n");
    // Initialize buffer maps
    buffInitialize(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix);
    //printMatrix(h_bufferMatrix, NUM_PEERS, NUM_BLOCKS);

    // create timer
    cutResetTimer(timer);
    cutStartTimer(timer);

    // Disseminating blocks using GRF
    cpuDelayGRF = p2pGRF(NUM_PEERS, NUM_BLOCKS, h_bufferMatrix, h_neighbourList);

    // stop timer
    cutStopTimer(timer);
    CPUtotaltime = cutGetTimerValue(timer);

    printf("CPU: GRF Rounds = %d, time = %.6f s\n", cpuDelayGRF, CPUtotaltime/1000);
    printf("CPU Processing Speed: %.6f s/round\n", CPUtotaltime/1000/cpuDelayGRF);
    printf("GPU Speedup: %.6f\n", CPUtotaltime/cpuDelayGRF/GPUtotaltime*gpuDelayGRF);
    //========= memory clean up ==========
    //printf("Deallocating CPU/GPU memory ...\n");
    
    myFree(h_neighbourList);
    myFree(h_bufferMatrix);
    myFree(h_bufUpdate);
    myFree(h_bufCount);

    CUT_SAFE_CALL( cutDeleteTimer( timer));

    //check if kernel execution generated and error
    CUT_CHECK_ERROR("Kernel execution failed");
    //CUT_EXIT(argc, argv);

    return(0);
}
